/*
* dbstruct.h


Copyright (C) 2008-2021 Alessandro Vesely

This file is part of Ipqbdb.

Ipqbdb is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Ipqbdb is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DBSTRUCT_H_INCLUDED
#define DBSTRUCT_H_INCLUDED

#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <db.h>

#define IPQBDB_CHK_SIGNATURE 0X5a
#define IPBQDB_VERSION 1

/*
* This is the data part of the main database
* (the key part is the IPv4 address, uint32_t, in network order)
* optimized assuming sizeof: time_t=8, int=4, double=8, pid_t=4
*/
typedef struct ip_data_t
{
	double decay;          // >= IPQBDB_NEVER_DECAY for always blocked
	time_t last_update, last_block, created;
	int probability;       // 0-RAND_MAX probability of being blocked
	int reason_id;         // index into the reason file
	int block_cnt;         // number of packets blocked
	int decay_add_cnt;     // number of threshold upward passes
// --- added since v.0
	int caught_cnt;        // number of times caught (from ban or parse)
	uint32_t start_range;  // reserved --must be 0
	uint16_t rep_ruling;   // added in v. 1.05
	unsigned char flag;    // reserved --must be 0
	unsigned char nu[4];   // reserved --must be 0
	unsigned char chk;     // CHK_SIGNATURE, for loose type checking
} ip_data_t;

/*
* This is the data part of the "bynumber" (primary) description table
* (the key part is the reason_id number)
* The last field of this structure is a variable length string containing
* the textual description, and is the secondary key "bydescr" of the table.
*/
typedef struct bynumber_descr_t
{
	uint64_t caught_cnt, rec_cnt,  // set on block record creation/update
		block_cnt, decay_add_cnt;  // set on block record deletion
	time_t created,                // this record creation
		last_added,                // set on block record creation/update
		last_removed;              // set on block record deletion
	unsigned char report_category; // external category added in v. 1.07
	unsigned char nu[6];           // reserved --must be 0
	unsigned char chk;             // CHK_SIGNATURE, for loose type checking
	char descr[/*IPQBDB_DESCR_MAX_LENGTH*/];
} bynumber_descr_t;

/*
* This is the data part of the whitelist
* (the key part is the IPv4 address)
*/
typedef struct ip_white_t
{
	double decay;	 // 0 = never block this ip, otherwise set initial decay
	int probability;      // reserved --must be 0
	int reason_id;        // reserved --must be 0
	uint32_t start_range; // reserved --must be 0
	unsigned char nu[11]; // reserved --must be 0
	unsigned char chk;    // CHK_SIGNATURE, for loose type checking
} ip_white_t;

// used to be here
#include "report_error.h"

void fputs_database_help(void);
char *database_fname(char const *name, app_private*);
char *database_fname_with_home(char const *name, app_private *ap,
	char const *home, int add_slash_if);
int open_database(char *name, app_private *ap, DB_ENV**, DB**);
int close_db(DB *db);
int close_dbenv(DB_ENV *db_env, int cleanup);


enum
{ get_descr_quoted = 1, get_descr_add_id = 2, get_descr_add_category = 16 };
int get_descr(DB *db, int id, int opt, char *dest, size_t size);
int last_descr_ndx(DB *db1);
int get_reason_id(int *reason_id, char *reason_string,
	app_private *ap, DB *db1, DB *db2, int verbose);
enum
{ set_desc_old_record = 0, set_descr_new_record = 1 };
int set_descr_cnt(DB *db1, int descr_id, int is_new_record,
	ip_data_t const *deldata, time_t const *tloc);
int open_descrdb(char *fname, DB_ENV *db_env, DB** db1, DB** db2);
int check_whitelist(DB *, uint32_t ip, double *decay);

typedef void (*write_block_cb)(uint32_t ip_addr, int reason_id,
	int write_block_force,
	ip_data_t *ip_data, ip_data_t *old_data, DB *db_d);
enum
{ write_block_force_decay = 1, write_block_force_probability = 2,
	write_block_force_reason = 4,
	write_block_force_add_category = 16 };
int write_block(DB *db, uint32_t ip_addr, double decay, int reason_id,
	int* /* I/O */ probability, int force, write_block_cb verbose_fn, DB *db_d);

void fputs_initial_count_help(void);

#endif // DBSTRUCT_H_INCLUDED
